///////////////////////////////////////////////////
// ADSInterrupt.cpp
// 720020-1149x
//
// Applied Data Systems
//
// See ADSInterrupt.h for history and notes.
//
///////////////////////////////////////////////////

#include "ADSInterrupt.h"

#define SYSINTR_NOP					0x00
#define	SA1110_GPIO_BASE			0x90040000
#define	SA1110_GPDR_OFFSET			2
#define	MEMORY_SIZE					0x10000 
#define DIRECTION_INPUT				0
#define DIRECTION_OUTPUT			1

// VirtualCopy definition
extern "C" BOOL VirtualCopy(LPVOID lpvDestMem, LPVOID lpvSrcMem, 
							DWORD dwSizeInBytes, DWORD dwProtectFlag);

///////////////////////////////////////////////////
// ADSInterrupt
// Class Constructor
///////////////////////////////////////////////////
ADSInterrupt::ADSInterrupt()
{
	m_hBceddkDll = NULL;
	m_hCoreDll = NULL;
	m_bInit = FALSE;
	m_bConnected = FALSE;
	m_hInterrupt = SYSINTR_NOP;
	m_hIntEvent = 0;

	// Load the libraries needed
	DEBUGMSG(TRUE, (_T("Loading bceddk.dll...")));
	m_hBceddkDll = LoadLibrary(_T("bceddk.dll"));

	if( m_hBceddkDll == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}

	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), m_hBceddkDll));
	DEBUGMSG(TRUE, (_T("Loading coredll.dll...")));

	m_hCoreDll = LoadLibrary((L"coredll.dll"));
	if( m_hCoreDll == NULL)	
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), m_hCoreDll));

	DEBUGMSG(TRUE, (_T("Binding APIs...\r\n")));

	/////////////// InterruptConnect ////////////////
	DEBUGMSG(TRUE, (_T("InterruptConnect =")));
	InterruptConnect = (INTERRUPTCONNECT)GetProcAddress(m_hBceddkDll,(L"InterruptConnect"));	
	if(InterruptConnect == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), InterruptConnect));

	/////////////// InterruptDisconnect ////////////////
	DEBUGMSG(TRUE, (_T("InterruptDisconnect =")));
	InterruptDisconnect = (INTERRUPTDISCONNECT)GetProcAddress(m_hBceddkDll,(L"InterruptDisconnect"));	
	if(InterruptDisconnect == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), InterruptDisconnect));

	/////////////// InterruptInitialize ////////////////
	DEBUGMSG(TRUE, (_T("InterruptInitialize =")));
	InterruptInitialize = (INTERRUPTINITIALIZE)GetProcAddress(m_hCoreDll,(L"InterruptInitialize"));	
	if(InterruptInitialize == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), InterruptInitialize));

	/////////////// InterruptDisable ////////////////
	DEBUGMSG(TRUE, (_T("InterruptDisable =")));
	InterruptDisable = (INTERRUPTDISABLE)GetProcAddress(m_hCoreDll,(L"InterruptDisable"));	
	if(InterruptDisable == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), InterruptDisable));

	/////////////// InterruptDone ////////////////
	DEBUGMSG(TRUE, (_T("InterruptDone =")));
	InterruptDone = (INTERRUPTDONE)GetProcAddress(m_hCoreDll,(L"InterruptDone"));
	if(InterruptDone == NULL)
	{
		DEBUGMSG(TRUE, (_T("FAILED\r\n")));
		return;
	}
	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), InterruptDone));
}

///////////////////////////////////////////////////
// ~ADSInterrupt
// Class Destructor
///////////////////////////////////////////////////
ADSInterrupt::~ADSInterrupt()
{
	// clean up
	this->Disable();  // internally calls Done()
	this->Disconnect();

	// free libraries
	if(m_hBceddkDll != NULL)
		FreeLibrary(m_hBceddkDll);

	if(m_hCoreDll != NULL)
		FreeLibrary(m_hCoreDll);
}

///////////////////////////////////////////////////
// Connect
// This method must be called before any other method
// in the class.
// [I] InterruptNumber - Interrupt number to connect
//     These are defined in ADSInterrupt.h
// [I] InterruptMode - this is the mode in which to connect
//     This includes level or edge triggering as well
//     as which edge (rising/falling)
// [O] BOOL - TRUE is successful, FALSE if not
///////////////////////////////////////////////////
BOOL ADSInterrupt::Connect(ULONG InterruptNumber, ULONG InterruptMode)
{
	DEBUGMSG(TRUE, (_T("Connecting interrupt %i..."), InterruptNumber));

	// if a GPIO was requested, ensure it is set for input
	switch(InterruptNumber)
	{
		case SA1100_GPIO25_EDGE:
			SetGPIODirection(25, DIRECTION_INPUT);
			break;
		case SA1100_GPIO26_EDGE:
			SetGPIODirection(26, DIRECTION_INPUT);
			break;
		case SA1100_GPIO27_EDGE:
			SetGPIODirection(27, DIRECTION_INPUT);
			break;
	}

	m_hInterrupt = InterruptConnect(Internal,
							        0,
							        InterruptNumber,
									InterruptMode);
	if(m_hInterrupt == SYSINTR_NOP)
	{
		DEBUGMSG(TRUE, (_T("FAILED: InterruptConnect returned SYSINTR_NOP\r\n"), GetLastError()));
		m_bConnected = FALSE;
		return FALSE;
	}

	DEBUGMSG(TRUE, (_T("0x%08x\r\n"), m_hInterrupt));
	
	m_bConnected = TRUE;

	return TRUE;
}

///////////////////////////////////////////////////
// Initialize
// Call this after Connect and before any other method
// [I] EventToSet - Handle of an event.  When the 
//     interrupt occurs, this event will be set
//     These are defined in ADSInterrupt.h
// [O] long - error information or zero for no error
///////////////////////////////////////////////////
long ADSInterrupt::Initialize(HANDLE EventToSet)
{
	DEBUGMSG(TRUE, (_T("Initializing interrupt...")));

	if(!m_bConnected)
	{
		DEBUGMSG(TRUE, (_T("FAILED - must call InterruptConnect first\r\n")));
		return -1;
	}
	
	if(EventToSet == 0)
	{
		DEBUGMSG(TRUE, (_T("FAILED - invalid event handle\r\n")));
		return -2;
	}

	m_hIntEvent = EventToSet;

	BOOL bret = InterruptInitialize(m_hInterrupt,
    	    						m_hIntEvent,
    	    						NULL,
    	    						0);

    if(!bret)
	{
		DEBUGMSG(TRUE, (_T("FAILED (%d)\r\n"), GetLastError()));
		return GetLastError();
    }

	DEBUGMSG(TRUE, (_T("done\r\n")));
	
	m_bInit = TRUE;

	return 0;
}

///////////////////////////////////////////////////
// Disable
// Call this to remove our event from the Interrupt
// [I] none
// [O] none
///////////////////////////////////////////////////
void ADSInterrupt::Disable()
{
	DEBUGMSG(TRUE, (_T("Disabling interrupt...")));

	if(!m_bConnected)
	{
		DEBUGMSG(TRUE, (_T("FAILED - no interrupt is connected\r\n")));
		return;
	}

	InterruptDisable(m_hInterrupt);

	DEBUGMSG(TRUE, (_T("done\r\n")));
}

///////////////////////////////////////////////////
// Done
// Call this after handling an event to re-enable
// the interrupt.  The interrupt can not fire again
// for any thread waiting on it until this is called
// [I] none
// [O] long - error information or zero for no error
///////////////////////////////////////////////////
long ADSInterrupt::Done()
{
	DEBUGMSG(TRUE, (_T("Finishing interrupt...")));
	
	if(!m_bConnected)
	{
		DEBUGMSG(TRUE, (_T("FAILED - no interrupt is connected\r\n")));
		return -1;
	}

	InterruptDone(m_hInterrupt);

	DEBUGMSG(TRUE, (_T("done\r\n")));

	return 0;
}

///////////////////////////////////////////////////
// Disconnect
// Disconnect our event (from Initialize) from the
// interrupt.
// [I] none
// [O] long - error information or zero for no error
///////////////////////////////////////////////////
long ADSInterrupt::Disconnect()
{
	DEBUGMSG(TRUE, (_T("Disconnecting interrupt...")));

	if(!m_bConnected)
	{
		DEBUGMSG(TRUE, (_T("FAILED - no interrupt is connected\r\n")));
		return -1;
	}

	// first make sure it's marked as handled
	this->Done();

	// now disconnect
	InterruptDisconnect(m_hInterrupt);

	
	DEBUGMSG(TRUE, (_T("done\r\n")));

	return 0;
}

///////////////////////////////////////////////////
// SetGPIODirection
// Set the direction of the GPDR register
// interrupt.
// [I] none
// [O] long - error information or zero for no error
///////////////////////////////////////////////////
void ADSInterrupt::SetGPIODirection(int GPIO, BYTE Direction)
{
	LPVOID pAddr;
	DWORD *pMem;

	pAddr = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
	
	if(pAddr)	
	{
		if(VirtualCopy (pAddr, (PVOID)(SA1110_GPIO_BASE >> 8), 
				MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
				DEBUGMSG(1, (_T("\r\n0x%08x mapped to: 0x%08x\r\n"), SA1110_GPIO_BASE, pAddr));
		}
		else 
		{
			RETAILMSG(1, (_T("Error! at VirtualCopy()\r\n")));

			return;
		}
	}
	else 
	{
		RETAILMSG(1, (_T("Error! at VirtualAlloc()\r\n")));
		return;
	}

	pMem = (DWORD*)pAddr;
	DWORD mask = 0;

	// set the register
	DEBUGMSG(TRUE, (_T("\r\n    GPDR old: 0x%08x\r\n"), pMem[SA1110_GPDR_OFFSET / sizeof(DWORD)]));
	if(Direction != 0)
	{
		mask = (1 << GPIO);
		pMem[SA1110_GPDR_OFFSET / sizeof(DWORD)] |= mask;
	}
	else
	{
		mask = ~(1 << GPIO);
		pMem[SA1110_GPDR_OFFSET / sizeof(DWORD)] &= mask;
	}

	DEBUGMSG(TRUE, (_T("    masked with: 0x%08x\r\n"), mask));
	DEBUGMSG(TRUE, (_T("    GPDR new: 0x%08x\r\n"), pMem[SA1110_GPDR_OFFSET / sizeof(DWORD)]));

	VirtualFree(pAddr, 0, MEM_RELEASE);
}